@@ -28,7 +28,7 @@
   :use-module (srfi srfi-35)
   :use-module (srfi srfi-39)
   :use-module (r6rs bytevector)
-  :autoload   (r6rs i/o ports)      (lookahead-u8)
+  :autoload   (r6rs io ports)       (lookahead-u8)
 
   :export (procedure-call-information rpc-call?
            rpc-call-xid rpc-call-program rpc-call-version
@@ -41,20 +41,23 @@
            rpc-program-version-lookup-procedure
            rpc-procedure-number rpc-procedure-handler
            rpc-procedure-argument-xdr-type rpc-procedure-result-xdr-type
+           rpc-procedure-one-way?
 
-           lookup-called-procedure handle-procedure-call
-           handle-procedure-lookup-error
+           lookup-called-procedure
 
            make-i/o-manager i/o-manager?
            i/o-manager-exception-handler i/o-manager-read-handler
            run-input-event-loop
 
            make-rpc-listening-socket-i/o-manager
-           serve-one-stream-request run-stream-rpc-server
+           serve-one-stream-request serve-one-stream-request/asynchronous
+           handle-procedure-call handle-procedure-lookup-error
+           run-stream-rpc-server
            current-stream-connection stream-connection?
            stream-connection-port stream-connection-peer-address
            stream-connection-rpc-program
 
+           ;; deprecated
            serve-one-tcp-request run-tcp-rpc-server
            current-tcp-connection tcp-connection?
            tcp-connection-port tcp-connection-peer-address
@@ -184,12 +187,21 @@ call message, an error condition is raised."
   (lookup-procedure %rpc-program-version-lookup-procedure))
 
 (define-record-type <rpc-procedure>
-  (make-rpc-procedure number argument-xdr-type result-xdr-type handler)
+  (%make-rpc-procedure number argument-xdr-type result-xdr-type handler
+                       one-way?)
   rpc-procedure?
   (number             rpc-procedure-number)
   (argument-xdr-type  rpc-procedure-argument-xdr-type)
   (result-xdr-type    rpc-procedure-result-xdr-type)
-  (handler            rpc-procedure-handler))
+  (handler            rpc-procedure-handler)
+  (one-way?           rpc-procedure-one-way?))
+
+(define (make-rpc-procedure number argument-xdr-type result-xdr-type handler
+                            . rest)
+  (%make-rpc-procedure number argument-xdr-type result-xdr-type handler
+                       (if (null? rest)
+                           #f
+                           (car rest))))
 
 
 (define (compile-numbered-items get-number items)
@@ -302,6 +314,15 @@ the appropriate reply message via @var{send-result}."
            (xdr-encode! raw-reply 0 rpc-message reply)
            (send-result raw-reply 0 reply-size)))))
 
+
+
+(define %rpc-reply-message-size
+  ;; Pre-compute the size of a successful reply message.
+  (xdr-type-size rpc-message
+                 (make-rpc-message 0
+                                   'REPLY 'MSG_ACCEPTED
+                                   'SUCCESS)))
+
 (define (handle-procedure-call call programs input-port send-result)
   "Handle procedure call @var{call} using @var{programs}, reading input from
 @var{input-port} and sending the result via @var{send-result}, a
@@ -313,21 +334,58 @@ count."
     (let* ((proc           (lookup-called-procedure call programs))
            (handler        (rpc-procedure-handler proc))
            (input-type     (rpc-procedure-argument-xdr-type proc))
-           (result         (handler (xdr-decode input-type input-port)))
-           (result-type    (rpc-procedure-result-xdr-type proc))
-           (result-size    (xdr-type-size result-type result))
+           (result         (handler (xdr-decode input-type input-port))))
 
-           (reply-prologue (make-rpc-message (rpc-call-xid call)
-                                             'REPLY 'MSG_ACCEPTED
-                                             'SUCCESS))
-           (prologue-size  (xdr-type-size rpc-message reply-prologue))
-           (total-size     (+ result-size prologue-size))
-           (raw-result     (make-bytevector total-size)))
-      (xdr-encode! raw-result
-                   (xdr-encode! raw-result 0
-                                rpc-message reply-prologue)
-                   result-type result)
-      (send-result raw-result 0 total-size))))
+      (if (not (rpc-procedure-one-way? proc))
+          (let* ((result-type    (rpc-procedure-result-xdr-type proc))
+                 (result-size    (xdr-type-size result-type result))
+
+                 (reply-prologue (make-rpc-message (rpc-call-xid call)
+                                                   'REPLY 'MSG_ACCEPTED
+                                                   'SUCCESS))
+                 (total-size     (+ result-size %rpc-reply-message-size))
+                 (raw-result     (make-bytevector total-size)))
+            (xdr-encode! raw-result
+                         (xdr-encode! raw-result 0
+                                      rpc-message reply-prologue)
+                         result-type result)
+            (send-result raw-result 0 total-size))))))
+
+(define (handle-procedure-call/asynchronous call programs input-port
+                                            send-result)
+  "Handle procedure call @var{call} using @var{programs}, reading input from
+@var{input-port} and sending the result via @var{send-result}, a
+three-argument procedure that is passed a bytevector, offset and octet
+count.  Unlike with @code{handle-procedure-call}, the procedure from
+@var{programs} that is called is passed a continuation argument that it must
+invoke explicitly with the value to return to the client."
+  (define (make-result-handler result-type)
+    (lambda (result)
+      (let* ((result-size    (xdr-type-size result-type result))
+
+             (reply-prologue (make-rpc-message (rpc-call-xid call)
+                                               'REPLY 'MSG_ACCEPTED
+                                               'SUCCESS))
+             (total-size     (+ result-size %rpc-reply-message-size))
+             (raw-result     (make-bytevector total-size)))
+        (xdr-encode! raw-result
+                     (xdr-encode! raw-result 0
+                                  rpc-message reply-prologue)
+                     result-type result)
+        (send-result raw-result 0 total-size))))
+
+
+  (guard (c ((rpc-procedure-lookup-error? c)
+             (handle-procedure-lookup-error c call send-result)))
+
+    (let* ((proc           (lookup-called-procedure call programs))
+           (handler        (rpc-procedure-handler proc))
+           (input-type     (rpc-procedure-argument-xdr-type proc)))
+      (handler (xdr-decode input-type input-port)
+               (if (rpc-procedure-one-way? proc)
+                   (lambda (result) result)
+                   (make-result-handler
+                    (rpc-procedure-result-xdr-type proc)))))))
 
 
 ;;;
@@ -420,27 +478,37 @@ is left."
 ;;; High-level aids.
 ;;;
 
-(define (serve-one-stream-request program port)
-  "Serve one RPC for @var{program}, reading the RPC from @var{port} (using
-the record-marking protocol) and writing the reply to @var{port}.  If
-@var{port} is closed or the end-of-file was reached, an
-@code{&rpc-connection-lost-error} is raised."
-  (let ((input-port (rpc-record-marking-input-port port)))
+(define (make-stream-request-server handle-call)
+  (lambda (program port)
+    ;; Serve one RPC for @var{program}, reading the RPC from @var{port}
+    ;; (using the record-marking protocol) and writing the reply to
+    ;; @var{port}.  If @var{port} is closed or the end-of-file was reached,
+    ;; an @code{&rpc-connection-lost-error} is raised."
+    (let ((input-port (rpc-record-marking-input-port port)))
+
+      (catch 'system-error
+        (lambda ()
+          (if (eof-object? (lookahead-u8 port))
+              (raise (condition (&rpc-connection-lost-error
+                                 (port port))))
+              (let* ((msg  (xdr-decode rpc-message input-port))
+                     (call (procedure-call-information msg)))
+                (handle-call call (list program) input-port
+                             (lambda (bv offset count)
+                               (send-rpc-record port bv
+                                                offset count))))))
+        (lambda (key . args)
+          (raise (condition (&rpc-connection-lost-error
+                             (port port)))))))))
+
+(define serve-one-stream-request
+  ;; The synchronous variant.
+  (make-stream-request-server handle-procedure-call))
+
+(define serve-one-stream-request/asynchronous
+  ;; The asynchronous, continuation-passing style RPC server.
+  (make-stream-request-server handle-procedure-call/asynchronous))
 
-    (catch 'system-error
-      (lambda ()
-        (if (eof-object? (lookahead-u8 port))
-            (raise (condition (&rpc-connection-lost-error
-                               (port port))))
-            (let* ((msg  (xdr-decode rpc-message input-port))
-                   (call (procedure-call-information msg)))
-              (handle-procedure-call call (list program) input-port
-                                     (lambda (bv offset count)
-                                       (send-rpc-record port bv
-                                                        offset count))))))
-      (lambda (key . args)
-        (raise (condition (&rpc-connection-lost-error
-                           (port port))))))))
 
 (define-record-type <stream-connection>
   (make-stream-connection port address rpc-program)
