@@ -1,5 +1,5 @@
 ;;; GNU Guile-RPC --- A Scheme implementation of ONC RPC.
-;;; Copyright (C) 2007  Free Software Foundation, Inc.
+;;; Copyright (C) 2007, 2008  Free Software Foundation, Inc.
 ;;;
 ;;; This file is part of GNU Guile-RPC.
 ;;;
@@ -17,6 +17,7 @@
 ;;; along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 (define-module (rpc xdr)
+  :autoload   (srfi srfi-1)    (fold)
   :use-module (srfi srfi-9)
   :autoload   (srfi srfi-34)   (raise)
   :use-module (srfi srfi-35)
@@ -212,12 +213,10 @@ type."
 ;;; Type size.
 ;;;
 
-(define (array-length vec)
-  "Return the length of a one-dimensional array (e.g., a vector) or @code{#f}
-if @var{vec} is not a one-dimensional array."
-  (let ((dim (array-dimensions vec)))
-    (and (null? (cdr dim))
-         (car dim))))
+(define array-length generalized-vector-length)
+(define array-ref    generalized-vector-ref)
+(define array-set!   generalized-vector-set!)
+
 
 (define (xdr-type-static-size type)
   "If @var{type} has a fixed size (once encoded), known statically, i.e.,
@@ -281,6 +280,10 @@ independently of the value of type @var{type} being encoded, then return it
              (+ 4 (loop (xdr-union-arm-type type discr)
                         (cdr value)))))
 
+          ((procedure? type)
+           ;; delayed type
+           (loop (type) value))
+
           (else
            (raise (condition
                    (&xdr-unknown-type-error (type type))))))))
@@ -358,13 +361,17 @@ independently of the value of type @var{type} being encoded, then return it
            (let liip ((types  (xdr-struct-base-types type))
                       (values value)
                       (index  index))
-             (if (null? values)
-                 (round-up-size index)
-                 (liip (cdr types)
-                       (cdr values)
-                       (encode (car types)
-                               (car values)
-                               index)))))
+             (cond ((null? types)
+                    (round-up-size index))
+                   ((null? values)
+                    ;; not enough struct fields provided
+                    (type-error type value))
+                   (else
+                    (liip (cdr types)
+                          (cdr values)
+                          (encode (car types)
+                                  (car values)
+                                  index))))))
 
           ((xdr-union-type? type)
            (let ((discr (car value))
@@ -373,6 +380,10 @@ independently of the value of type @var{type} being encoded, then return it
              ;; We can safely assume that the discriminant is 32-bit.
              (encode (xdr-union-arm-type type discr) arm (+ index 4))))
 
+          ((procedure? type)
+           ;; delayed type
+           (encode (type) value index))
+
           (else
            (raise (condition (&xdr-unknown-type-error (type type))))))))
 
@@ -400,12 +411,12 @@ independently of the value of type @var{type} being encoded, then return it
              (decode type port)))
 
           ((xdr-vector-type? type)
-           (let* ((max    (xdr-vector-max-element-count type))
-                  (type   (xdr-vector-base-type type))
-                  (decode (and (xdr-basic-type? type)
-                               (xdr-basic-type-vector-decoder type)))
-                  (raw    (get-bytevector-n port 4))
-                  (len    (bytevector-u32-ref raw 0 %xdr-endianness)))
+           (let* ((max     (xdr-vector-max-element-count type))
+                  (type    (xdr-vector-base-type type))
+                  (vdecode (and (xdr-basic-type? type)
+                                (xdr-basic-type-vector-decoder type)))
+                  (raw     (get-bytevector-n port 4))
+                  (len     (bytevector-u32-ref raw 0 %xdr-endianness)))
 
              (if (and max (> len max))
                  (raise (condition
@@ -414,8 +425,8 @@ independently of the value of type @var{type} being encoded, then return it
                           (element-count len))))
                  (let ((padding (vector-padding type len))
                        (result
-                        (if decode
-                            (decode type len port)
+                        (if vdecode
+                            (vdecode type len port)
                             (let ((vec (make-vector len)))
                               (let liip ((index 0))
                                 (if (< index len)
@@ -428,21 +439,20 @@ independently of the value of type @var{type} being encoded, then return it
                    result))))
 
           ((xdr-struct-type? type)
-           (let liip ((types  (xdr-struct-base-types type))
-                      (result '()))
-             (if (null? types)
-                 (let ((padding (or (xdr-struct-padding type) 0)))
-                   (if (> padding 0) (get-bytevector-n port padding))
-                   (reverse! result))
-                 (let ((value (decode (car types))))
-                   (liip (cdr types)
-                         (cons value result))))))
+           (let ((padding (or (xdr-struct-padding type) 0))
+                 (result  (map decode (xdr-struct-base-types type))))
+             (if (> padding 0) (get-bytevector-n port padding))
+             result))
 
           ((xdr-union-type? type)
            (let* ((discr (decode (xdr-union-discriminant-type type)))
                   (value (decode (xdr-union-arm-type type discr))))
              (cons discr value)))
 
+          ((procedure? type)
+           ;; delayed type
+           (decode (type)))
+
           (else
            (raise (condition (&xdr-unknown-type-error (type type))))))))
 
